#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the aiuna project.
#  Please respect the license - more about this in the section (*) below.
#
#  aiuna is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  aiuna is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with aiuna.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.


import os
import signal
from contextlib import contextmanager

import time


class withTiming:
    @staticmethod
    def cpu():
        """CPU time.

        Returns
        -------
            Sum of all SO times except wall time.
        """
        #  return time.process_time()  # Does not include children processes.
        t = os.times()
        return t[0] + t[1] + t[2] + t[3]

    @staticmethod
    def time(f):
        start = withTiming.cpu()
        ret = f()
        end = withTiming.cpu()
        return end - start, ret

    @staticmethod
    def clock():
        """Wall clock time.

        Returns
        -------
            Ellapsed time.
        """
        return time.monotonic()

    @staticmethod
    @contextmanager
    def time_limit(seconds=None):
        if seconds is None:
            yield
        else:
            def signal_handler(signum, frame):
                raise TimeoutException("Timed out!")

            signal.signal(signal.SIGALRM, signal_handler)
            signal.alarm(seconds)
            try:
                yield
            finally:
                signal.alarm(0)


class TimeoutException(Exception):
    pass



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the aiuna project.
#  Please respect the license - more about this in the section (*) below.
#
#  aiuna is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  aiuna is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with aiuna.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.
#
#  aiuna is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  aiuna is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with aiuna.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.


import inspect
import json
from functools import cached_property
from aiuna.content.root import Root
from aiuna.content.creation import read_arff, new
from aiuna.step.new import New
from akangatu.transf.dataindependentstep_ import DataIndependentStep_


class File(DataIndependentStep_):
    def __init__(self, name="iris.arff", path="./", hashes=None):
        self.isclass = False
        if not path.endswith("/"):
            print(path)
            raise Exception("Path should end with '/'", path)
        if name.endswith(".arff"):
            self._partial_config = {"name": name, "path": path}
            self.filename = path + name
        else:
            raise Exception("Unrecognized file extension:", name)
        self._hashes = hashes
        super().__init__(config_func=self._config)

    def _process_(self, data):
        if data is not Root:
            raise Exception(f"{self.name} only accepts Root as Data. Use File(  ).data or Sink instead.", type(data))
        return self.data

    def _config(self):
        config = self._partial_config
        config["hashes"] = self.hashes
        return config

    def _uuid_(self):  # override uuid because default uuid is based on config, which includes unreliable args
        return New(hashes=self.hashes).uuid

    # REMINDER File cannot be lazy, since it depends on file content to know the uuid which will be used to generate
    # the lazy Data object.
    ###@cached_property
    @property
    def data(self):
        d = read_arff(self.filename)
        ds = d["dataset"], d["description"], d["matrices"], d["original_hashes"]
        self._dataset, self._description, matrices, original_hashes = ds
        if self._hashes:
            if self._hashes != original_hashes:
                raise Exception(
                    f"Provided hashes f{self._hashes} differs from hashes of file content: " f"{original_hashes}!")
        else:
            self._hashes = original_hashes

        return new(**matrices)

    ###@cached_property
    @property
    def dataset(self):
        if self._hashes is None:
            _ = self.data  # force file reading
        return self._dataset

    ###@cached_property
    @property
    def description(self):
        if self._hashes is None:
            _ = self.data  # force file reading
        return self._description

    ###@cached_property
    @property
    def hashes(self):
        if self._hashes is None:
            _ = self.data  # force calculation of hashes
        return self._hashes

    def parameters(self):
        return self._config()

    def __call__(self, *args, **kwargs):
        return File(*args, **kwargs)

    # def _context_(self):
    #     return "aiuna.step.file"


file = File()
# TODO: autocomplete vai funcionar para args dos callable?

# TODO: forçar implementação do __call__ em vez de __init__?



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the aiuna project.
#  Please respect the license - more about this in the section (*) below.
#
#  aiuna is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  aiuna is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with aiuna.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.


from functools import cached_property

from garoupa.uuid import UUID

from aiuna.leaf import Leaf
from akangatu.transf.mixin.printing import withPrinting


class History(withPrinting):
    isleaf = False

    def __init__(self, step=None, nested=None, uuid=None):
        """Optimized iterable "list" of Leafs (wrapper for a step or a dict) based on structural sharing."""
        if step:
            if nested:
                raise Exception("History cannot be built from both 'step' and 'nested'!")
            self._last = step
            self.uuid = step.uuid
            self.nested = [Leaf(step)]
        elif nested:
            if uuid is None:
                raise Exception("History cannot be built from 'nested' without 'uuid'!")
            self._last = nested[-1].last
            self.uuid = uuid
            self.nested = nested
        else:
            self._last = None
            self.uuid = UUID.identity
            self.nested = []

    @property
    def id(self):
        return self.uuid.id

    @property
    def last(self):
        if self._last is None:
            raise Exception("Empty history has no last item.")
        return self._last

    ###@cached_property
    @property
    def aslist(self):
        return [step.asdict for step in self]

    def _asdict_(self):
        return {step.id: step.desc for step in self}

    def __add__(self, other):
        return History(nested=[self, other], uuid=self.uuid * other.uuid)

    def __lshift__(self, step):
        h = History(step)
        return History(nested=[self, h], uuid=self.uuid * h.uuid)

    def traverse(self, node):
        # TODO: remove recursion due to python conservative limits for longer histories (AL, DStreams, ...)
        if node.isleaf:
            yield node.asstep
        else:
            for tup in node.nested:
                yield from self.traverse(tup)

    def __iter__(self):
        yield from self.traverse(self)

    def __xor__(self, attrname):
        """Shortcut ^ to get an attribute along all steps."""
        # touch properties to avoid problems (is this really needed?)
        # void = [a.name + a.longname for a in self.traverse(self)]   #TODO voltar essa linha?
        return list(map(lambda x: getattr(x, attrname), self.traverse(self)))

    def __eq__(self, other):
        return self.uuid == other.uuid



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the aiuna project.
#  Please respect the license - more about this in the section (*) below.
#
#  aiuna is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  aiuna is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with aiuna.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.


import _pickle as pickle
import json
from functools import lru_cache
from random import random

import lz4.frame as lz
import numpy as np
import zstandard as zs
# Things that should be calculated only once.
# ##################################################
from akangatu.transf.config import safety
# TODO: make a permanent representative dictionary and check if it
#  reduces compression time and size of textual info like transformations.
from garoupa.encoders import integers2bytes, bytes2integers


@lru_cache()
def compression_dict():  # TODO: update dictionary
    txt_comming_from_a_file_for_training_of_compressor = "aasfdkjgd kajgsdf af kajgakjsg fkajgs df\n" * 653
    samples = [
        (str(random()) + s).encode()
        for s in txt_comming_from_a_file_for_training_of_compressor.split("\n")
    ]
    return zs.train_dictionary(dict_size=99999999, samples=samples, threads=-1)


# ##################################################

class HashableBinary:
    def __init__(self, n, obj):
        self.n = n
        self.obj = obj

    def __hash__(self):
        return self.n


def fpack(data, field):
    return memopack(HashableBinary(data.uuids[field].n, data[field]))


# ###@lru_cache()
def memopack(hashable_binary):
    return pack(hashable_binary.obj)


# compression_lock = new_lock()


def pack(obj):
    cctx = zs.ZstdCompressor(threads=-1)
    cctxdic = zs.ZstdCompressor(threads=-1, dict_data=compression_dict(), write_dict_id=False)

    cctxdec = zs.ZstdDecompressor()
    cctxdicdec = zs.ZstdDecompressor(dict_data=compression_dict())

    # with safety():
    if isinstance(obj, np.ndarray) and str(obj.dtype) == "float64" and len(obj.shape) == 2:
        # X, ...
        h, w = obj.shape
        fast_reduced = lz.compress(obj.reshape(w * h), compression_level=1)
        header = integers2bytes(obj.shape)
        return b"F" + header + cctx.compress(fast_reduced)
    elif isinstance(obj, (list, set, str, int, float, bytearray, bool)):
        # steps, ...
        js = json.dumps(obj, sort_keys=True, ensure_ascii=False)
        return b"J" + cctx.compress(js.encode())
    elif isinstance(obj, str):
        return b"T" + cctxdic.compress(obj.encode())  # b'T'+0s==1409286144
    elif isinstance(obj, bytes):
        fast_reduced = lz.compress(obj, compression_level=1)
        return b"B" + cctx.compress(fast_reduced)
    elif callable(obj):
        raise Exception("Cannot compress callable", type(obj))
    else:
        # print(f"Unknown type {type(obj)}{'-' + str(obj.dtype) if 'ndarray' in str(type(obj)) else ''} to compress, using pickle...")
        # categorical ndarrays, ...
        pickled = pickle.dumps(obj)  # 1169_airlines explodes here with RAM < ?
        fast_reduced = lz.compress(pickled, compression_level=1)
        return b"P" + cctx.compress(fast_reduced)  # b'P'+0s==1342177280


def unpack(dump_with_header):
    cctx = zs.ZstdCompressor(threads=-1)
    cctxdic = zs.ZstdCompressor(threads=-1, dict_data=compression_dict(), write_dict_id=False)

    cctxdec = zs.ZstdDecompressor()
    cctxdicdec = zs.ZstdDecompressor(dict_data=compression_dict())

    # with safety():
    header = dump_with_header[:1]
    dump = dump_with_header[1:]
    if header == b"P":
        decompressed = lz.decompress(cctxdec.decompress(dump))
        return pickle.loads(decompressed)
    elif header == b"T":
        return cctxdicdec.decompress(dump).decode()
    elif header == b"B":
        return lz.decompress(cctxdec.decompress(dump))
    elif header == b"F":
        header = dump_with_header[1:9]
        dump = dump_with_header[9:]
        decompressed = lz.decompress(cctxdec.decompress(dump))
        [h, w] = bytes2integers(header)
        return np.reshape(np.frombuffer(decompressed), newshape=(h, w))
    elif header == b"J":
        return json.loads(cctxdec.decompress(dump).decode())
    else:
        raise Exception("Unknown compression format:", header, dump[:300])

# def pack_object(obj):  #blosc is buggy
#     """
#     Nondeterministic (fast) parallel compression!
#     Due to multithreading, blosc is nondeterministic and useless for UUIDs.
#     :param obj:
#     :return:
#     """
#     import blosc
#     pickled = pickle.dumps(obj)
#     fast_reduced = lz.compress(pickled, compression_level=1)
#     return blosc.compress(fast_reduced,
#                       shuffle=blosc.NOSHUFFLE, cname='zstd', clevel=3)


# def unpack_object(dump):  #blosc is buggy
#     import blosc
#     decompressed = blosc.decompress(dump)
#     fast_decompressed = lz.decompress(decompressed)
#     return pickle.loads(fast_decompressed)



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the aiuna project.
#  Please respect the license - more about this in the section (*) below.
#
#  aiuna is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  aiuna is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with aiuna.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.
from unittest import TestCase

from aiuna.history import History
from garoupa.uuid import UUID

from akangatu.operator.nullary.empty import SingletonException
from akangatu.transf.noop import NoOp


class Test(TestCase):
    def test_storage(self):
        from aiuna.content.root import Root
        self.assertTrue(Root >> NoOp == Root)
        self.assertRaises(SingletonException, lambda: Root({}, UUID(), History(NoOp)))
        self.assertTrue(Root >> NoOp == Root)



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the aiuna project.
#  Please respect the license - more about this in the section (*) below.
#
#  aiuna is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  aiuna is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with aiuna.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.
from akangatu.transf.dataindependentstep_ import DataIndependentStep_
import numpy as np


class Let(DataIndependentStep_):
    def __init__(self, field, value):
        super().__init__(field=field, value=value)
        self.field = field
        if isinstance(value, (list, np.ndarray)):
            self.value = value
        elif isinstance(value, (float, int)):
            self.value = np.array([[value]])
        else:
            raise Exception("Unknown value type", type(value))

    # TODO accept and store any kind of value in Data, not only 2d ndarrays (and str lists)

    # TODO if it is not jsonable or it is larger than 250 (utf8) chars,
    #    raises exception saying to use Set instead, which will md5-hash it

    def _process_(self, data):
        dic = {self.field: lambda: self.value}
        return data.update(self, **dic)



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the aiuna project.
#  Please respect the license - more about this in the section (*) below.
#
#  aiuna is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  aiuna is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with aiuna.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

import traceback
from functools import lru_cache, cached_property
from pprint import pprint

import arff
import numpy as np
from pandas import DataFrame, Series

from aiuna.content.creation import new, translate_type
from aiuna.mixin.timing import withTiming, TimeoutException
from garoupa.avatar23 import colors
from garoupa.uuid import UUID
from akangatu.linalghelper import evolve_id, mat2vec, field_as_matrix, islazy
from akangatu.transf.mixin.identification import withIdentification
from akangatu.transf.mixin.printing import withPrinting
from akangatu.transf.noop import NoOp
from akangatu.transf.step import Step, MissingField
from akangatu.transf.timeout import Timeout


class Data(withIdentification, withPrinting, withTiming):
    """Immutable lazy data for most machine learning scenarios.
# comparable: Fields precedence when comparing which data is greater.
    Parameters
    ----------
    history
        A History objects that represents a sequence of steps.
    storage_info
        An alias to a global Storage object for lazy matrix fetching.
    matrices  TODO trocar pra fields
        A dictionary like {X: <numpy array>, Y: <numpy array>}.
        Matrix names should have a single uppercase character, e.g.:
        X=[
           [23.2, 35.3, 'white'],
           [87.0, 52.7, 'brown']
        ]
        Y=[
           'rabbit',
           'mouse'
        ]
        They can be, ideally, numpy arrays (e.g. storing is optimized).
        A matrix name followed by a 'd' indicates its description, e.g.:
        Xd=['weight', 'height', 'color']
        Yd=['class']
        A matrix name followed by a 't' indicates its types ('ord', 'int',
        'real', 'cat'*).
        * -> A cathegorical/nominal type is given as a list of nominal values:
        Xt=['real', 'real', ['white', 'brown']]
        Yt=[['rabbit', 'mouse']]

        Volatile fields, i.e., those that can be lost due to their nondeterministic nature,
        should be marked with the mutability suffix '_' like in 'duration_';
        so they stay outside matrix, but still available as Data attributes.
        They are not stored by conventional storages.
        User-content enabled ones like OkaSt may provide that.
    """

    # muda em                           update  iniget  iniget  iniget  init    step.new()      get     update
    #       Let         Let                 c/ Tim.                         vários  map,cache,summ
    #       maxtime  comparable  changed failure timeout duratio parntid inner   stream  storage step    history
    # updatekwargs:x        x
    # protegid/automat:                 x       x       x       x       x                       x       nonkw   x
    # transformavel x       x                                                   x       x
    # ____________________________________________________________________________________________________________________
    # initargs                          x       x       x       x       calculado               x       x
    # _fields_m:                                                x
    # ____________________________________________________________________________________________________________________
    # updtargs:                                                                 x       x               x
    # armazen:  field       f           f       f       f       run     col     col     s+bool          x
    # não lazy: x           x           [dispara lazies?              ] x               it/dblz x       x

    # REMINDER todos fields entram obrig lazy no update() porque podem ativar o consumo de stream antes da hora.
    # REMINDER tudo lazy adotado inicialmente pq ajuda o cache a saber o que vai ser feito (e usuario a puxar do DB) sem precisar mexer com uuid
    # REMINDER changed precisa vir do update() pq não sabemos se algum lazy veio do step anterior
    triggers = ["failure", "timeout", "duration"]
    maxtime, comparable = None, None
    _duration = 0

    def __hash__(self):
        return id(self)

    # def __hash__(self):
    #     return self.uuid.n  # não sei pq , mas não tá herdando hash de withIdentification

    def __init__(self, uuid, uuids, history, **fields):
        if "changed" not in fields:
            print(uuids.keys())
            print(fields.keys())
            raise Exception("Field 'changed' is mandatory as a kwarg, alongside its UUID inside uuids.")
        self.changed = fields["changed"]
        self.field_funcs_m = fields
        self._uuid, self.uuids = uuid, uuids
        try:
            step = history.last
        except:
            step = NoOp()
        self.step_func_m = step
        # lazy lambda name format: "_stepuuid..._from_storage_storageuuid..."
        self.step_uuid = step.uuid if isinstance(step, Step) else UUID(step.name[1:24])
        self.parent_uuid = uuid / self.step_uuid
        self.history = history

        # TODO: Check if types (e.g. Mt) are compatible with values (e.g. M).
        # TODO lembrar por que concluí que "stream and inner shold also be lazy"

    @property
    def step(self):
        if islazy(self.step_func_m):
            self.step_func_m = self.step_func_m()
        return self.step_func_m

    ###@cached_property
    @property
    def changed_asdict(self):
        """Evaluate and return all fields in 'changed'."""
        return {lazy_field: self.field_funcs_m[lazy_field] for lazy_field in self.changed}

    @property
    def failure(self):
        """First touch all fields in 'changed' (which are necessarily lazy), and only then return the failure, if any."""
        _ = self.changed_asdict
        return self._failure

    @property
    def timeout(self):
        """First touch all fields in 'changed' (which are necessarily lazy), and only then return whether there was a timeout."""
        _ = self.changed_asdict
        return self._timeout

    @property
    def duration(self):
        """First touch all fields in 'changed' (which are necessarily lazy), and only then return duration value."""
        _ = self.changed_asdict
        return self._duration
        # TODO como duration viria do storage?

    @property
    def hasinner(self):
        return "inner" in self.field_funcs_m

    @property
    def hasstream(self):
        return "stream" in self.field_funcs_m

    def _asdict_(self):
        raise Exception("Should not be accessed. Use asdict property instead!")

    # TODO  verificar se algo imutavel depende do asdict (que vai ser mutavel)
    @property
    def asdict(self):  # overriding because there are lazy fields in Data, so asdict is mutable and noncached
        dic = {"uuid": self.uuid, "uuids": self.uuids,
               "step": self.step_func_m.name if islazy(self.step_func_m) else self.step_func_m.asdict}
        if self.hasinner:
            dic["inner"] = self.field_funcs_m["inner"].__name__ if islazy(self.inner) else self.inner.asdict
        if self.hasstream:
            dic["stream"] = "iterator"
        dic.update(self.field_funcs_m)
        return dic

    def inners(self):
        data = self
        while True:
            yield data
            if not data.hasinner:
                break
            data = data.inner

    def update(self, step, **fields):
        """Recreate an updated Data object.

        Parameters
        ----------
        steps
            Step object that process this Data object.
        fields
            Matrices or vector/scalar shortcuts to them.

        Returns
        -------
        New Data object (it keeps references to the old one for performance).
        :param step:
        """
        if not isinstance(step, Step):
            raise Exception("Step cannot be of type", type(step))
        if isinstance(step, Timeout):
            changed = ["timeout", "duration"]
        else:
            from aiuna.step.file import File
            from aiuna.step.new import New
            changed = []
            for field, value in fields.items():
                if not islazy(value) and not isinstance(step, (File, New)):
                    raise Exception(f"{field} should be callable! Not:", type(value))
                if field in self.triggers + ["changed"]:
                    raise Exception(f"'{field}' cannot be externally set! Step:" + step.longname)
                changed.append(field)
        changed = list(sorted(changed))

        # Only update field uuids for the provided and changed ones!
        updated_fields = {"changed": changed}
        # REMINDER: conversão saiu de _update() para self[] (que é o novo .field()) para conciliar laziness e aceitação de vetores e escalares
        for k, v in fields.items():
            kup = k.upper() if len(k) == 1 else k
            if (kup not in self.field_funcs_m) or self.field_funcs_m[kup] is not v:
                updated_fields[kup] = v
        uuid, uuids = evolve_id(self.uuid, self.uuids, step, updated_fields)

        newfields = self.field_funcs_m.copy()
        newfields.update(updated_fields)
        return Data(uuid, uuids, self.history << step, **newfields)

    ###@cached_property
    @property
    def eager(self):
        """Touch every lazy field by accessing all fields.

        Stream is kept intact???"""
        for f in self:
            _ = self[f]
        return self

    ###@cached_property
    @property
    def Xy(self):
        return self["X"], self["y"]

    @property
    def uuid(self):
        return self._uuid

    @property
    def id(self):
        return self._uuid.id

    @property
    def _uuid_(self):
        return self._uuid

    def __lt__(self, other):
        """Amenity to ease pipeline result comparisons. 'A > B' means A is better than B."""
        if self.comparable is None:
            return Exception("Impossible to make comparisons. No comparable field was defined!")
        for name in self.comparable:
            if name in self.field_funcs_m:
                return self[name] < other[name]
        return Exception("Impossible to make comparisons. None of the comparable fields are available:",
                         self.comparable)

    def __eq__(self, other):
        # TODO benchmark presence of isinstance here, can be disabled in production version
        return isinstance(other, Data) and self.uuid == other.uuid

    # REMINDER Data is mutable in rare occasions: setitem/delitem/

    ###@lru_cache
    def arff(self, relation, description):
        Xt = [untranslate_type(typ) for typ in self.Xt]
        Yt = [untranslate_type(typ) for typ in self.Yt]
        dic = {
            "description": description,
            "relation": relation,
            "attributes": list(zip(self.Xd, Xt)) + list(zip(self.Yd, Yt)),
            "data": np.column_stack((self.X, self.Y)),
        }
        try:
            return arff.dumps(dic)
        except:
            traceback.print_exc()
            print("Problems creating ARFF")
            print("Types:", Xt, Yt)
            print("Sample:", self.X[0], self.Y[0])
            print("Expected sizes:", len(Xt), "+", len(Yt))
            print("Real sizes:", len(self.X[0]), "+", len(self.Y[0].shape))
            exit(0)

    # def __rlshift__(self, other):
    #     if other.isclass:
    #         other = other()
    #     return other.process(self)

    def __rshift__(self, other):
        if other.isclass:
            other = other()
        return other.process(self)

    def __add__(self, other):
        """Merge two Data objects"""
        # se mexeram na mesma matriz, vale a segunda.
        # Dá pra emendar com o histórico da segunda, desde que o uuid final seja recalculado e que
        # operações de ajuste (Copy) preservem as matrizes pedidas pela segunda no inicio da divergencia (normalmente após File/New).
        # Assume que haver matrizes extra não altera resultado de um step (o que é sensato, pois step não deve fazer "inspection")
        # se foi misto,
        raise Exception("Not implemented: should solve conflicts to merge history of two data objects.")

    def mutate(self, newdata):
        """Inplace swapping of this Data object by another

        UUID and __hash__ will change."""
        self._uuid = newdata.uuid
        self.uuids = newdata.uuids
        self.step_func_m = newdata.step_func_m
        self.step_uuid = newdata.step_uuid
        self.parent_uuid = newdata.parent_uuid
        self.field_funcs_m = newdata.field_funcs_m
        self.history = newdata.history

        # TODO: use this instead?
        # self.__dict__ = newdata.__dict__

        # # REMINDER: cache invalidation seems unneeded according to tests
        # for attr in self.__dict__.values():
        #     if callable(attr) and hasattr(attr, "cacheclear"):
        #         attr.cacheclear()

    ###@lru_cache()
    def __getitem__(self, key):
        """Safe access to a field, with a friendly error message.

        Parameters
        ----------
        name
            Name of the field.

        Returns
        -------
        Matrix, vector or scalar
        """
        # Format single-letter field according to capitalization.
        if len(key) == 1:
            kup = key.upper()
            if key.islower():
                return mat2vec(self[kup])
        else:
            kup = key

        # Is it an already evaluated field?
        if not islazy(self.field_funcs_m[kup]):
            return self.field_funcs_m[kup]

        # Is it a lazy field...
        #   ...from storage? Just call it, without timing or catching exceptions as failures.
        if "_from_storage_" in self.field_funcs_m[kup].__name__:
            self.field_funcs_m[kup] = field_as_matrix(key, self.field_funcs_m[kup]())
            return self.field_funcs_m[kup]

        #   ...yet to be processed?
        try:
            with self.time_limit(self.maxtime):
                t, value = self.time(lambda: field_as_matrix(key, self.field_funcs_m[kup]()))
                self._duration += t
                self.field_funcs_m[kup] = value
        except TimeoutException:
            self.mutate(self >> Timeout(self.maxtime))
        except Exception as e:
            self._failure = self.step.translate(e, self)
            self.field_funcs_m[kup] = None  # REMINDER None means interrupted
        return self.field_funcs_m[kup]

    def __getattr__(self, item):
        """Create shortcuts to fields."""

        if item.startswith("__"):
            return super().__getattribute__(item)

        # Handle field access
        if item in self.field_funcs_m or item.upper() in self.field_funcs_m:
            return self[item]

        # Handle missing known fields
        if item in ["inner", "stream"]:
            raise MissingField(f"Field '{item}' not provided by steps" + str(self.history ^ "name"), self, item)

        # Fallback to methods and other attributes.
        return super().__getattribute__(item)

    def __delitem__(self, key):
        from aiuna.step.delete import Del
        self.mutate(self >> Del(field=key))

    def __setitem__(self, key, value):
        # process mutation
        from aiuna.step.let import Let
        self.mutate(self >> Let(field=key, value=value))

    def items(self):
        # TODO quais ficam de fora de fields? são importantes pra iterar?
        for k in self:
            yield k, self[k]

    def __iter__(self):
        yield from self.field_funcs_m

    def __len__(self):
        return len(self.field_funcs_m)  # TODO idem

    def _name_(self):
        return self._uuid

    def _context_(self):
        return self.history ^ "names"

    @staticmethod
    def from_pandas(X_pd: DataFrame, y_pd: Series):
        X, y = X_pd.to_numpy(), y_pd.to_numpy().astype("float")
        Xd, Yd = X_pd.columns.tolist(), [y_pd.name]
        Xt, Yt = [translate_type(str(c)) for c in X_pd.dtypes], list(sorted(set(y)))
        return new(X=X, y=y, Xd=Xd, Yd=Yd, Xt=Xt, Yt=Yt)

    # @property
    # ###@lru_cache()
    # def ids_lst(self):  #TODO ainda precisa?
    #     return [self.uuids[name].id for name in self.names]
    #
    # @property
    # ###@lru_cache()
    # def ids_str(self):  #TODO ainda precisa?
    #     return ','.join(self.ids_lst)
    #
    # @property
    # ###@lru_cache()
    # def names_str(self):  #TODO ainda precisa?
    #     return ','.join(self.names)

    # ###@cached_property
    @property
    # def picklable(self):
    #     """Remove unpicklable parts."""
    #     return self.picklable_()[0]
    #
    # ###@cached_property
    @property
    # def unpicklable(self):
    #     """Restore unpicklable parts."""
    #     return self.unpicklable_([{}])

    # # noinspection PyDefaultArgument
    # def picklable_(self, unpicklable_parts=[]):
    #     """Remove unpicklable parts, but return them together as a list of dicts, one dict for each nested inner objects."""
    #     if isinstance(self, Picklable):
    #     return self, unpicklable_parts
    #     unpicklable_parts = unpicklable_parts.copy()
    #     history = History(self.history.aslist)
    #     unpicklable_parts.append({"stream": self.stream})
    #     if self.inner:
    #     inner, unpicklable_parts = self.inner.picklable_(unpicklable_parts)
    #     else:
    #     inner = None
    #     newdata = Picklable(self.uuid, self.uuids, history, stream=None, storage_info=self.storage_info, inner=inner, **self._matrices)
    #     # TODO: como ficam os lazies antes de picklear?
    #     #  no fetch blz, só falta usar a lista de unpicklables
    #     #  pra recuperar após passar pela thread.
    #     #  Mas proibe dar store? ou store tenta ativar todos?
    #     return newdata.eager, unpicklable_parts

    # def unpicklable_(self, unpicklable_parts):
    #     """Rebuild an unpicklable Data.
    #
    #     History is desserialized, if given as str.
    #     """
    #     # REMINDER: Persistence/threading actions can be nested, so self can be already unpicklable
    #     if not isinstance(self, Picklable):
    #     return self
    #     if self.stream:
    #     raise Exception("Inconsistency: this picklable Data object contains a stream!")
    #     # make a copy, since we will change history and stream directly; and convert to right class
    #     stream = unpicklable_parts and "stream" in unpicklable_parts[0] and unpicklable_parts[0]["stream"]
    #     if not isinstance(self.history[0], dict):
    #     raise Exception("Pickable Data should have a History of dicts instead of", type(self.history[0]))
    #     inner = self.inner and self.inner.unpicklable
    #     return Data(self.uuid, self.uuids, self.history, stream, self.storage_info, inner, **self._matrices)

    @property
    def icon(self):
        """Information to create an icon for this Data object."""
        return {"id": self.id, "step": self.step.asdict_rec, "colors": colors(self.id)}

    # ###@cached_property
    @property
    def past(self):
        """Map ancestor_id -> afterstep/icon_colors. Last item refers to the current Data object."""
        from aiuna.content.root import Root
        dic = {}
        d = Root
        for s in list(self.history) + [NoOp()]:
            dic[d.id] = {"step": s.asdict_rec, "colors": colors(d.id)}
            d >>= s
        return dic


def untranslate_type(name):
    if isinstance(name, list):
        return name
    if name in ["real", "int"]:
        return "NUMERIC"
    raise Exception("Unknown type:", name)


""" importante TODO

-({a0, a1} + {b0, b1}) = a0 | a1 | + {b0, b1} 


    step =    an infinitely nested unit set (INUS)    A = {{ ... {a} ... }}; or,
              a set of steps                          D = {A, B, C}

    A * B                     product            (** cached product)
    A + B                     union
    A | B                     stack
    +S                        summation (union) of all unit sets in stack S
    -A                        stacking (disunion) of all unit sets in set A

    Properties (A: set, S: stack, a: INUS)
    +A = error
    -S = error
    -a = a

    -(A | B) = +A | +B
    s0 | s1 | s2              sequence s0,s1,s2
    s0 | ...                  sequence s0,s0,...,s0
    +S                        sequence s0,s1,...,sn
    +(A|B) = {a0,a1,b0,b1}
    -{a0,a1,b0,b1} = (A|B)

    ~                         sample
    ~~                        ?

    @ cache
    & | ^ // %
    -step -> hold
    ~step -> sample
     ...
    +step -> permite ser destrutivo?

    d.hash * A.hash * B.hash / e.hash
    s.id
    update() deveria aceitar qq obj com: {hash, id ?} ou aceitar hexmd5?

    ###########################################
    aceitar repetições de step, melhorar hash ou forçar sanduiches de step recheados com algo inerte?
    coisas a considerar no hash:
          embaralhamento?
          AB != BA
          AAAAA sem colisão
          AA != I
    ###########################################

    ------------------------
    def alternativa, que enfraquece o uso dos ops unários:
    A unit set is equivalent to its contained element.
    A stack of a single element is the element itself.
    ------------------------
"""



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the aiuna project.
#  Please respect the license - more about this in the section (*) below.
#
#  aiuna is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  aiuna is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with aiuna.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.
#
#  aiuna is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  aiuna is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with aiuna.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.


import json
import math
from hashlib import md5

# noinspection PyUnresolvedReferences
import arff
import numpy as np
import sklearn.datasets as ds

from akangatu.linalghelper import fields2matrices


def mathash(k, v):
    try:
        dump = json.dumps(v, sort_keys=True, ensure_ascii=False).encode() if isinstance(v, list) else v.tobytes()
        return md5(dump).hexdigest()
    except TypeError as e:
        print(f"Cannot calculate hash for {k} with value {v}")
        exit()


def hashes_mats(fields):
    """Calculate pseudo-unique hash for each field."""
    matrices = dict(fields2matrices(fields).items())
    return {k: mathash(k, v) for k, v in matrices.items()}, matrices


def new(**fields):
    from aiuna.step.new import New
    hashes, matrices = hashes_mats(fields)
    return New(hashes, **matrices).data


# noinspection PyPep8Naming
def read_arff(filename):
    """
    Create  from ARFF file.

    Assume X,y classification task and last attribute as target.

    Parameters
    ----------
    filename
        path of the dataset

    Returns
    -------
    (dict of matrix hashes and metainfo)
    """
    # Load file.
    file = open(filename, "r")
    dic = arff.load(file, encode_nominal=False)  # ['description', 'relation', 'attributes', 'data']
    name = dic["relation"]
    description = dic["description"]
    file.close()

    # Extract attributes and targets.
    Arr = np.array(dic["data"])
    Att = dic["attributes"][0:-1]
    TgtAtt = dic["attributes"][-1]

    # Extract X values (numeric when possible), descriptions and types.
    X = Arr[:, 0:-1]
    Xd = [tup[0] for tup in Att]
    Xt = [translate_type(tup[1]) for tup in Att]
    if len(nominal_idxs(Xt)) == 0:
        X = X.astype(float)

    # Extract Y values (assumes categorical), descriptions and types.
    Y = np.ascontiguousarray(Arr[:, -1].reshape((Arr.shape[0], 1)))
    Yd = [TgtAtt[0]]
    Yt = [translate_type(TgtAtt[1])]

    original_hashes, matrices = hashes_mats({"X": X, "Y": Y, "Xd": Xd, "Yd": Yd, "Xt": Xt, "Yt": Yt})
    return {"dataset": name, "description": description, "matrices": matrices, "original_hashes": original_hashes}


def translate_type(name):
    if isinstance(name, list):
        return name
    name = name.lower()
    if name in ["numeric", "real"] or name.startswith("float"):
        return "real"
    elif name.startswith("int"):
        return "int"
    else:
        raise Exception("Unknown type:", name)


def read_csv(filename, target='class'):
    raise NotImplementedError
    # df = pd.read_csv(filename)  # 1169_airlines explodes here with RAM < 6GiB
    # return read_data_frame(df, filename, target)


def random_classification_dataset(n_attributes, n_classes, n_instances):
    """
    Generate a syntetic Data object for a classification task.
    :param n_attributes:
    :param n_classes:
    :param n_instances:
    :return:
    """
    n = int(math.sqrt(2 * n_classes))
    X, y = ds.make_classification(
        n_samples=n_instances, n_features=n_attributes, n_classes=n_classes, n_informative=n + 1
    )
    raise NotImplementedError
    # name = "RndData-" + uuid(pickle.dumps((X, y)))
    # dataset = Dataset(name, "rnd", X=enumerate(n_attributes * ["rnd"]), Y=["class"])
    # return Data(dataset, X=X, Y=as_column_vector(y))


def as_column_vector(vec):
    return vec.reshape(len(vec), 1)


def nominal_idxs(M):
    return [idx for idx, val in list(enumerate(M)) if isinstance(val, list)]



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the aiuna project.
#  Please respect the license - more about this in the section (*) below.
#
#  aiuna is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  aiuna is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with aiuna.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.


from akangatu.transf.step import Step


class Leaf:
    """Contains a step or a dict, answers for both."""
    isleaf, _dict, _step = True, None, None

    def __init__(self, step):
        if isinstance(step, dict):
            self._dict = step
        else:
            self._step = step

    @property
    def asstep(self):
        if self._step is None:
            self._step = Step.fromdict(self._dict)
        # self._step = self._step() if lazy(self._step) else self._step
        return self._step

    @property
    def asdict(self):
        if self._dict is None:
            self._dict = self._step.asdict
        return self._dict

    def __getattr__(self, item):
        if item in ["asstep", "asdict"]:
            return super().__getattribute__(item)
        return getattr(self.asstep, item)

    def __getitem__(self, item):
        return self.asdict[item]

    def _asdict_(self):
        return self._dict



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the aiuna project.
#  Please respect the license - more about this in the section (*) below.
#
#  aiuna is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  aiuna is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with aiuna.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.


# import traceback
#
#
# class asExceptionHandler:
#     def checked(self, f, exit_on_error):
#     """Error is different from failure."""
#
#     def func(data):
#         try:
#             return f(data)
#         except Exception as e:
#             msg = str(e)
#             if "n_co, n_features)=" in msg:
#                 pass  # We can improve msgs here, or just pass it as failure as it is.
#             else:
#                 # Handle errors, i.e. unpredictable failures.
#                 if "type object 'Root' has no attribute" in msg:
#                     msg = "Root data is the origin of all data and has no information: " + msg
#                 if exit_on_error:
#                     raise e
#                     # traceback.print_tb(e.__traceback__)
#                     # print(msg)
#                     # exit()
#             return data.failed(self, msg)
#
#     return func
#
# # numpy.warnings.filterwarnings("ignore")
# # numpy.warnings.filterwarnings("always")
